package br.com.dreamsource.mobile.components;

import br.com.dreamsource.mobile.jmesql.exceptions.SQLException;
import br.com.dreamsource.mobile.jmesql.jdbcResultSet;
import br.com.dreamsource.mobile.worders.WOrders;
import br.com.dreamsource.mobile.worders.forms.core.DatabaseManager;
import java.util.Vector;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.Item;

/**
 *
 * @author Elizeu Nogueira da Rosa Jr.
 * @version 0.1
 * @date 17.05.2007
 *
 */

public abstract class FormWOrders extends Form {
    
    protected WOrders midlet;
    protected CommandListener listener;

    protected Command nextPageCommand;
    protected Command previousPageCommand;
    protected Command insertCommand;
    protected Command editCommand;
    protected Command deleteCommand;
    protected Command saveCommand;
    protected Command cancelCommand;
    protected Command mainMenuCommand;
    
    protected Command locateCommand;
    protected Command backCommand;
    
    protected Command exitCommand;
    protected Command okCommand;
    
    protected Vector variables; 
    
    public FormWOrders(CommandListener listener, WOrders midlet, String title) {
        super(title, null);
        this.midlet = midlet;
        this.listener = listener;
        this.setCommandListener(listener);
        this.createComponents();        
        variables = new Vector(1);
    }
    
    protected Command getOkCommand() {
        if (okCommand == null) {
            okCommand = new Command("Ok", Command.OK, 1);
        }
        return okCommand;
    }    
    
    protected Command getExitCommand() {
        if (exitCommand == null) {
            exitCommand = new Command("Sair", Command.BACK, 1);
        }
        return exitCommand;
    }    
    
    protected Command getLocateCommand() {
        if (locateCommand == null) {
            locateCommand = new Command("Localizar", Command.SCREEN, 1);
        }
        return locateCommand;
    }    
    
    protected Command getBackCommand() {
        if (backCommand == null) {
            backCommand = new Command("Voltar", Command.BACK, 1);
        }
        return backCommand;
    }     
    
    protected Command getMainMenuCommand() {
        if (mainMenuCommand == null) {
            mainMenuCommand = new Command("Menu Principal", Command.SCREEN, 1);
        }
        return mainMenuCommand;
    }    
    
    protected Command getInsertCommand() {
        if (insertCommand == null) {
            insertCommand = new Command("Inserir", Command.SCREEN, 1);
        }
        return insertCommand;
    }     
    
    protected Command getEditCommand() {
        if (editCommand == null) {
            editCommand = new Command("Editar", Command.SCREEN, 1);
        }
        return editCommand;
    }    
    
    protected Command getDeleteCommand() {
        if (deleteCommand == null) {
            deleteCommand = new Command("Excluir", Command.SCREEN, 1);
        }
        return deleteCommand;
    }    
    
    protected Command getSaveCommand() {
        if (saveCommand == null) {
            saveCommand = new Command("Salvar", Command.OK, 1);
        }
        return saveCommand;
    }     
    
    protected Command getCancelCommand() {
        if (cancelCommand == null) {
            cancelCommand = new Command("Cancelar", Command.CANCEL, 1);
        }
        return cancelCommand;
    }     
    

    protected Command getNextPageCommand() {
        if (nextPageCommand == null) {
            nextPageCommand = new Command("Prxima", Command.SCREEN, 1);
        }
        return nextPageCommand;
    }    

    protected Command getPreviousPageCommand() {
        if (previousPageCommand == null) {
            previousPageCommand = new Command("Anterior", Command.SCREEN, 1);
        }
        return previousPageCommand;
    }
    
    /**
     * Mtodo chamado pelo contrutor para criar 
     * todos os componentes do form.
     */
    protected abstract void createComponents();
    
    /**
     * Mtodo invocado pelo midlet principal para 
     * todos os comandos
     */
    public abstract void fireCommand(Command command);
    
    protected void sendToForm(Form form) {
         Display.getDisplay(midlet).setCurrent(form);
    }
    
    public void setComponentValues(jdbcResultSet resultSet) throws SQLException {
        for (int i = 0; i < this.size(); i++) {
            Item item = this.get(i);
            if (item instanceof DBTextField) {
                ((DBTextField)item).setString(resultSet.getString(i + 1));
            } else
            if (item instanceof DBChoiceGroup) {
                //TODO verificar
            }
        }        
    }
    
    public void putVariable(String value) {
        this.variables.addElement(value);
    }
    
    public void removeVariable(String value) {
        this.variables.removeElement(value);
    }
    
    private Vector createVector() {
        Vector v = new Vector(this.size());
        int i = 0;
        for (i = 0; i < this.size(); i++) {
            Item item = this.get(i);
            if (item instanceof DBTextField) {
                v.addElement(((DBTextField)item).getString());
            } else
            if (item instanceof DBChoiceGroup) {
                v.addElement(((DBChoiceGroup)item).getSelectedItemID());
            }
        }        
        for (int j = i; j < variables.size() + i; j++) {
            v.addElement(variables.elementAt(j - i));
        }
        return v;
    }
    
    public boolean validateFields() {
        for (int i = 0; i < this.size(); i++) {
            Item item = this.get(i);
            if (item instanceof DBTextField) {
                if ((((DBTextField)item).getString().equals("")) && (! ((DBTextField)item).getNullable())) {
                    return false;
                }
            } else
            if (item instanceof DBChoiceGroup) {
                //v.addElement(((DBChoiceGroup)item).getSelectedItemID());
            }
        }        
        return true;
    }
    
    public void disableControls() {
        for (int i = 0; i < this.size(); i++) {
            Item item = this.get(i);
            if (item instanceof DBTextField) {
                ((DBTextField)item).setConstraints(DBTextField.UNEDITABLE);
            } else
            if (item instanceof DBChoiceGroup) {
                //((DBChoiceGroup)item).PLAIN
                //v.addElement(((DBChoiceGroup)item).getSelectedItemID());
            }
        }        
    }
    
    public void processUpdate(String sqlValue) throws SQLException {
        if (this.validateFields()) {
            DatabaseManager dbManager = new DatabaseManager(this.midlet.getUserPOJO());
            dbManager.update(sqlValue,  this.createVector());
        } else {
            throw new SQLException("Existem campos obrigatorios nao preenchidos!");
        }
    }
    
    public void processDelete(String sqlValue) throws SQLException {
        DatabaseManager dbManager = new DatabaseManager(this.midlet.getUserPOJO());
        dbManager.delete(sqlValue,  this.variables);
    }
    
    public void processInsert(String sqlValue) throws SQLException {
        if (this.validateFields()) {
            DatabaseManager dbManager = new DatabaseManager(this.midlet.getUserPOJO());
            dbManager.insert(sqlValue,  this.createVector());
        } else {
            throw new SQLException("Existem campos obrigatorios nao preenchidos!");
        }
    }
    
}
